
/***************************************************************************************************
                            
							  Lossless Raw Compressor

	File description
		Name:		Decoder.cpp

    Udpate Date:	2014.06.28

	Camera Platform Lab. 
	Digital Media R&D Center 
	Samsung Electronics Co., Ltd.

	Contact: wh0225.joo@samsung.com

***************************************************************************************************/

#include "define.h"
#include "Decoder.h"

//#define DUMP_PARSE_LOG

#ifdef DUMP_PARSE_LOG
FILE *fp_parse;
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////
// depend on BUS System
///////////////////////////////////////////////////////////////////////////////////////////////////

int32D5 getbits_64bit(DecoderParam *dp, int nBits);
int32D5 getbits_128bit(DecoderParam *dp, int nBits);
int32D5 getbits_sign_64bit(DecoderParam *dp, int nBits);
int32D5 getbits_sign_128bit(DecoderParam *dp, int nBits);

void SetBitstreamPos_64bit(DecoderParam *dp, int nLineNum);
void SetBitstreamPos_128bit(DecoderParam *dp, int nLineNum);

int32D5 (*getbits)(DecoderParam *, int);
int32D5 (*getbits_sign)(DecoderParam *, int);
void (*SetBitstreamPos)(DecoderParam *, int);

///////////////////////////////////////////////////////////////////////////////////////////////////
// Decoding Parameter
///////////////////////////////////////////////////////////////////////////////////////////////////

void IntializeDecoderParam(DecoderParam *dp)
{
	memset(dp, 0, sizeof(DecoderParam));
}

void bitstream_init(DecoderParam *dp)
{
	dp->nReadTotal	= 0;
	dp->bitstream	= dp->_bitstream;
	dp->nBitBuff	= 0;	
	dp->nBitPos		= (uint32D5)DATA_NOT_READY;				//  bitbuff data   ¸ ǹ
}

void InitDecoderSeq(DecoderParam *dp)
{
	dp->nWidth[0]		= dp->nOriWidth/2;
	dp->nWidth[0]		= ((dp->nWidth[0]+7)/8)*8;
	dp->nWidth[1]		= dp->nWidth[2]		= dp->nWidth[0];

	dp->nHeight[0]		= dp->nOriHeight;
	dp->nHeight[1]		= dp->nHeight[2]	= dp->nHeight[0]/2;
	dp->nBlockSize[0]	= dp->nBlockSize[1]	= dp->nBlockSize[2]	= 8;

	dp->nLineStartPos		= new uint32D5[dp->nHeight[0]];
	dp->nTileOffset			= new uint32D5[dp->nHeight[0]];
	dp->nTileBitPos			= new uint32D5[dp->nHeight[0]];
	dp->nTileBitBuff		= new uint64D5[dp->nHeight[0]];

	if(dp->bTileModeEnc==0)
		dp->nTileWidth	= dp->nWidth[0];
	dp->nOverlapWidth = 0;		// FEATURE: overlap not supported

	int i;
	for(i=0; i<3; i++)
		dp->pRec[i]		= new pixel[dp->nWidth[i]*dp->nHeight[i]*2];
}

void CloseDecoderSeq(DecoderParam *dp)
{
	delete dp->nLineStartPos;
	delete dp->nTileOffset;
	delete dp->nTileBitPos;
	delete dp->nTileBitBuff;

	delete dp->pRec[0];
	delete dp->pRec[1];
	delete dp->pRec[2];
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Image Header Parser
///////////////////////////////////////////////////////////////////////////////////////////////////

int ReadImageHeader(DecoderParam *dp)
{
	//12bit raw test
	//*
	//dp->Raw_stream_offset = 6023115;
	//dp->Raw_stream_size = 22963200;
	/**/
	//14bit raw test
	/*
	dp->Raw_stream_offset = 7045246;
	dp->Raw_stream_size = 29738848;
	/**/
	if(!dp->_bitstream)
		dp->_bitstream	= (uint32D5 *)malloc(sizeof(uint32D5) * dp->Raw_stream_size);

	memcpy(dp->_bitstream, &dp->srw_stream[dp->Raw_stream_offset], dp->Raw_stream_size);  

	dp->bitstream = dp->_bitstream;

	dp->nNLCVersion		= getbits_128bit(dp, 16);

	if(dp->nNLCVersion<0x5633) {		 // for backward compatibility

		getbits = getbits_64bit;
		getbits_sign = getbits_sign_64bit;
		SetBitstreamPos = SetBitstreamPos_64bit;

		dp->nImgFormat		= getbits(dp, 4);
		dp->nBitDepth		= getbits(dp, 4);
		dp->nOriWidth		= getbits(dp, 16);
		dp->nOriHeight		= getbits(dp, 16);
		dp->nNumBlkInRCUnit	= getbits(dp, 4);		// don't care	
		dp->nCompressionRatio = getbits(dp, 4);		// don't care
		dp->nTileWidth		= getbits(dp, 16);
		getbits(dp, 8);		// reserved
		dp->nOverlapWidth	= getbits(dp, 8);		// don't care
		getbits(dp, 8);		// reserved
		dp->nInc			= getbits(dp, 8);		// don't care
		getbits(dp, 2);		// reserved
		dp->nBnd_init_val	= getbits(dp, 14);

		dp->nOptCode = OPTIMIZE_SKIP | OPTIMIZE_MV | OPTIMIZE_QP;
		dp->bQPIncluded = 0;	// for SRW_V2.0, DRIMe4
		dp->nCompressionRatio = 11;
		dp->bAlwaysGFirst	= true;
	}
	else
	{
		getbits = getbits_128bit;
		getbits_sign = getbits_sign_128bit;
		SetBitstreamPos = SetBitstreamPos_128bit;

		dp->nImgFormat		= getbits(dp, 4);		
		dp->nBitDepth		= getbits(dp, 4);	
		dp->nNumBlkInRCUnit	= getbits(dp, 4);		// don't care	
		dp->nCompressionRatio = getbits(dp, 4);		// don't care
		dp->nOriWidth		= getbits(dp, 16);
		dp->nOriHeight		= getbits(dp, 16);
		dp->nTileWidth		= getbits(dp, 16);
		getbits(dp, 4);		// reserved
		dp->nOptCode		= getbits(dp, 4);
		dp->nOverlapWidth	= getbits(dp, 8);		// don't care
		getbits(dp, 8);		// reserved
		dp->nInc			= getbits(dp, 8);		// don't care
		getbits(dp, 2);		// reserved
		dp->nBnd_init_val	= getbits(dp, 14);

		dp->nBitDepth = dp->nBitDepth + 1;
		dp->bQPIncluded = (dp->nOptCode&0x4)>0 ? 0:1;
		//dp->nCompressionRatio = 11;
		dp->bAlwaysGFirst	= false;
	}

	//dp->nTileWidth		= dp->nTileWidth / 2;
	//dp->nOverlapWidth		= dp->nOverlapWidth / 2;

	dp->bTileModeEnc = 1;
	if(dp->nTileWidth==0 || dp->nTileWidth>=dp->nOriWidth)
		dp->bTileModeEnc = 0;
	//if(dp->bTileModeEnc==0)	dp->nTileWidth = 0;
	
	dp->bTileModeDec = 0;

#if 1
	int comp_code[12] = {0,10,20,25,30,40,50,60,70,80,90,100};
	DEBUG_PRINT("-----------------------------------------------------------\n");
	printf(" Ver. %4x   OPT %d   PW %d   PH %d   BD %d   CR %3d%%\n", 
			dp->nNLCVersion, dp->nOptCode, dp->nOriWidth, dp->nOriHeight, dp->nBitDepth, comp_code[dp->nCompressionRatio]);
	DEBUG_PRINT("-----------------------------------------------------------\n\n");
#endif

	//NX1 raw stream check.
	if (dp->nOriWidth == 6496 && dp->nOriHeight == 4336 && comp_code[dp->nCompressionRatio] == 100)	
		return true;
	else
		return false;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// File Dump
///////////////////////////////////////////////////////////////////////////////////////////////////

void ReconFrameDump(DecoderParam *dp)
{
	int32D5 _table[2][2] = {{0, 1}, {1, 0}};
	int32D5 *table;
	int32D5 i, j;
	pixel *pPtr;

	pPtr = new uint16D5[dp->nOriWidth * dp->nOriHeight * 3];

	if(dp->bAlwaysGFirst)
	{
		for(j=0; j<dp->nOriHeight; j+=2)
		{
			for(i=0; i<dp->nOriWidth; i+=2)
			{
				pPtr[(j+0)*dp->nOriWidth+(i+0)] = (dp->pRec[0][(j+0)*dp->nWidth[0]+(i/2)]);
				pPtr[(j+1)*dp->nOriWidth+(i+0)] = (dp->pRec[0][(j+1)*dp->nWidth[0]+(i/2)]);

				pPtr[(j+0)*dp->nOriWidth+(i+1)] = (dp->pRec[1][(j/2)*dp->nWidth[1]+(i/2)]);
				pPtr[(j+1)*dp->nOriWidth+(i+1)] = (dp->pRec[2][(j/2)*dp->nWidth[2]+(i/2)]);
			}
		}
	}
	else 
	{
		table = (((dp->nImgFormat>>2)&0x1)==((dp->nImgFormat>>3)&0x1)) ? _table[1] : _table[0];	// (GBRG) ? _table[0] : _table[1];
		for(j=0; j<dp->nOriHeight; j+=2)
		{
			for(i=0; i<dp->nOriWidth; i+=2)
			{
				pPtr[(j+0)*dp->nOriWidth+i+table[0]] = (dp->pRec[0][(j+0)*dp->nWidth[0]+(i/2)]);
				pPtr[(j+1)*dp->nOriWidth+i+table[1]] = (dp->pRec[0][(j+1)*dp->nWidth[0]+(i/2)]);

				pPtr[(j+0)*dp->nOriWidth+(i+table[1])] = (dp->pRec[1][(j/2)*dp->nWidth[1]+(i/2)]);
				pPtr[(j+1)*dp->nOriWidth+(i+table[0])] = (dp->pRec[2][(j/2)*dp->nWidth[2]+(i/2)]);
			}
		}
	}

	//dp.ip.fp_rec  = fopen(dp.ip.filename_rec [0], "wb");
	dp->ip.fp_rec  = fopen("dec_recon.raw", "wb");
	fwrite(pPtr, dp->nOriWidth * dp->nOriHeight, 2, dp->ip.fp_rec);
	fclose(dp->ip.fp_rec);

	delete pPtr;
}

void ReconFormatConvert(pixel *pDstBuff, DecoderParam *dp)	// Plane to Interleaved
{
	int32D5 _table[2][2] = {{0, 1}, {1, 0}};
	int32D5 *table;
	int32D5 i, j;

	if(dp->bAlwaysGFirst)
	{
		for(j=0; j<dp->nOriHeight; j+=2)
		{
			for(i=0; i<dp->nOriWidth; i+=2)
			{
				pDstBuff[(j+0)*dp->nOriWidth+(i+0)] = (dp->pRec[0][(j+0)*dp->nWidth[0]+(i/2)]);
				pDstBuff[(j+0)*dp->nOriWidth+(i+1)] = (dp->pRec[0][(j+1)*dp->nWidth[0]+(i/2)]);

				pDstBuff[(j+1)*dp->nOriWidth+(i+1)] = (dp->pRec[2][(j/2)*dp->nWidth[1]+(i/2)]);
				pDstBuff[(j+1)*dp->nOriWidth+(i+0)] = (dp->pRec[1][(j/2)*dp->nWidth[2]+(i/2)]);
			}
		}
	}
	else 
	{
		table = (((dp->nImgFormat>>2)&0x1)==((dp->nImgFormat>>3)&0x1)) ? _table[1] : _table[0];	// (GBRG) ? _table[0] : _table[1];
		for(j=0; j<dp->nOriHeight; j+=2)
		{
			for(i=0; i<dp->nOriWidth; i+=2)
			{
				pDstBuff[(j+0)*dp->nOriWidth+i+table[0]] = (dp->pRec[0][(j+0)*dp->nWidth[0]+(i/2)]);
				pDstBuff[(j+1)*dp->nOriWidth+i+table[1]] = (dp->pRec[0][(j+1)*dp->nWidth[0]+(i/2)]);

				pDstBuff[(j+0)*dp->nOriWidth+(i+table[1])] = (dp->pRec[1][(j/2)*dp->nWidth[1]+(i/2)]);
				pDstBuff[(j+1)*dp->nOriWidth+(i+table[0])] = (dp->pRec[2][(j/2)*dp->nWidth[2]+(i/2)]);
			}
		}
	}
}

void DumpRecon(unsigned char *pBuff, int nSize)
{
	FILE *fp = fopen("dec_reconUnPack.raw", "wb");
	if(fp==NULL) { printf("file open error : dec_recon.raw\n"); exit(-1); }
	fwrite(pBuff, nSize, 2, fp);
	fclose(fp);
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Util
///////////////////////////////////////////////////////////////////////////////////////////////////

void GetOffset(pixel *src[3], pixel *pos[3], int32D5 *nWidth, int32D5 x, int32D5 y)
{
	pos[0] = &src[0][y*nWidth[0]];			// G
	pos[1] = &src[1][(y/2)*nWidth[1]];		// 
	pos[2] = &src[2][(y/2)*nWidth[2]];		// 

	pos[0] += x;
	pos[1] += x;
	pos[2] += x;
}


void GetOffset2(pixel *src[3], pixel *pos[3], int32D5 *nCodedWidth, int32D5 x, int32D5 y, int32D5 nOffset)
{
	pos[0] = &src[0][(y+nOffset)*nCodedWidth[0]];
	pos[1] = &src[1][((y+nOffset*2)/2)*nCodedWidth[1]];
	pos[2] = &src[2][((y+nOffset*2)/2)*nCodedWidth[2]];

	pos[0] += x;
	pos[1] += x;
	pos[2] += x;
}


void GetReconPixel(DecoderParam *dp, pixel *pPred[3], int32D5 x, int32D5 y, int32D5 nOffset)
{
	pixel *pRef[3];
	int32D5 disp_i, disp_h;

	disp_i		= (nOffset+100)/2 - 50;		// integer-pel
	disp_h		= nOffset - 2*disp_i;		// half-pel
	
	GetOffset2(dp->pRec, pRef, dp->nWidth, x + disp_i, y, -1);

	int c, i;
	for(c=0; c<3; c++)
	{
		for(i=0; i<dp->nBlockSize[c]; i++)
		{
			pPred[c][i] = (disp_h) ? I_POL(pRef[c], i) : pRef[c][i];
		}
	}
}


void GetBlockSize(DecoderParam *dp, int y)
{
	dp->nBlockSize[0] = BLOCK_SIZE;
	dp->nBlockSize[1] = (y%2==0) ? BLOCK_SIZE : 0;
	dp->nBlockSize[2] = (y%2==0) ? 0 : BLOCK_SIZE;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Bit-stream
///////////////////////////////////////////////////////////////////////////////////////////////////

void SetBitstreamPos_64bit(DecoderParam *dp, int nLineNum)
{
  #ifdef D5_ELIMINATE_MMT
		if(nLineNum==0)
		{
			dp->bitstream	= (uint32D5*)((int)dp->_bitstream + 16);
		}
		else
		{
			int nLineOffset = ((uint32D5)dp->bitstream - (uint32D5)dp->_bitstream);
			int nPacketPos = (nLineOffset&0xf);
			dp->bitstream  = (nPacketPos==8 || nPacketPos==0) ? (uint32D5*)((int)dp->bitstream) : 
								(uint32D5*)((int)dp->bitstream - 4); 
		}
		
		dp->nBitPos		= (uint32D5)DATA_NOT_READY;
  #else
		dp->bitstream	= (uint32D5*)((int)dp->_bitstream + dp->nLineStart[nLineNum]);
		dp->nBitPos		= DATA_NOT_READY;
  #endif
}

void SetBitstreamPos_128bit(DecoderParam *dp, int nLineNum)
{
#ifdef D5_ELIMINATE_MMT
	int nLineOffset = ((size_t)dp->bitstream - (size_t)dp->_bitstream);
	int nPacketPos = (nLineOffset&0xf);

	dp->bitstream	= (nPacketPos==0) ? (uint32D5*)((size_t)dp->bitstream) :
					(nPacketPos==4) ? (uint32D5*)((size_t)dp->bitstream - 4) :
					(nPacketPos==8) ? ((dp->nBitPos>0) ? (uint32D5*)((size_t)dp->bitstream + 8) :(uint32D5*)((size_t)dp->bitstream - 8)) :
					(uint32D5*)((size_t)dp->bitstream + 4);
	dp->nBitPos		= (uint32D5)DATA_NOT_READY;
#else
	dp->bitstream	= (uint32D5*)((int)dp->_bitstream + dp->nLineStartPos[nLineNum]);
	dp->nBitPos		= (uint32D5)DATA_NOT_READY;
#endif
}

int32D5 getbits_64bit(DecoderParam *dp, int nBits)
{
	// ιƮ о ,  ó ǹǷ ׻ 2Ʈ ̻   ִٰ 
	if(dp->nBitPos==DATA_NOT_READY)
	{
		dp->nBitBuff = (((uint64D5)dp->bitstream[0])<<32) + (uint64D5)dp->bitstream[1];
		dp->bitstream +=2;
		dp->nBitPos = 0;
		dp->nReadTotal += 2;
	}

	int32D5 nVal = (int32D5)(dp->nBitBuff>>(64-nBits));
	dp->nBitBuff <<= nBits;
	dp->nBitPos += nBits;

	if(dp->nBitPos>=32)
	{
		dp->nReadTotal++;
		dp->nBitBuff += (((uint64D5)*dp->bitstream++)<<(dp->nBitPos-32));
		dp->nBitPos -= 32;
	}

	return nVal;
}

int32D5 getbits_128bit(DecoderParam *dp, int nBits)
{
	uint32D5 offset = ((size_t)dp->bitstream - (size_t)dp->_bitstream)/sizeof(uint32D5);

	// ιƮ о ,  ó ǹǷ ׻ 2Ʈ ̻   ִٰ 
	if((uint32D5)dp->nBitPos==(uint32D5)DATA_NOT_READY)
	{
		dp->nBitBuff = ((uint64D5)(dp->_bitstream[offset])<<32)+(uint64D5)(dp->_bitstream[offset+1]);
		dp->bitstream += 2;
		dp->nBitPos = 0;
		dp->nReadTotal += 2;
	}

	int32D5 nVal = (int32D5)(dp->nBitBuff>>(64-nBits));
	dp->nBitBuff <<= nBits;
	dp->nBitPos += nBits;
	
	if(dp->nBitPos>=32)
	{
		dp->nReadTotal++;
		dp->nBitBuff += (((uint64D5)dp->_bitstream[offset])<<(dp->nBitPos-32));
		dp->nBitPos -= 32;
		dp->bitstream++;
	}

	return nVal;
}


int32D5 getbits_sign_64bit(DecoderParam *dp, int nBits)
{
	// ιƮ о ,  ó ǹǷ ׻ 2Ʈ ̻   ִٰ 
	if(dp->nBitPos==DATA_NOT_READY)
	{
		dp->nReadTotal+=2;
		dp->nBitBuff = (((uint64D5)dp->bitstream[0])<<32) + (uint64D5)dp->bitstream[1];
		dp->bitstream+=2;
		dp->nBitPos = 0;
	}

	int32D5 nVal = (int32D5)(((int64D5)dp->nBitBuff)>>(64-nBits));
	dp->nBitBuff <<= nBits;
	dp->nBitPos += nBits;

	if(dp->nBitPos>=32)
	{
		dp->nReadTotal++;
		dp->nBitBuff += (((uint64D5)*dp->bitstream++)<<(dp->nBitPos-32));
		dp->nBitPos -= 32;
	}

	return nVal;
}

int32D5 getbits_sign_128bit(DecoderParam *dp, int nBits)
{
	uint32D5 offset = ((size_t)dp->bitstream - (size_t)dp->_bitstream)/sizeof(uint32D5);

	// ιƮ о ,  ó ǹǷ ׻ 2Ʈ ̻   ִٰ 
	if(dp->nBitPos==(uint32D5)DATA_NOT_READY)
	{
		dp->nReadTotal+=2;
		//dp->nBitBuff = (((uint64D5)dp->bitstream[0])<<32) + (uint64D5)dp->bitstream[1];
		dp->nBitBuff = (((uint64D5)dp->_bitstream[offset])<<32) + (uint64D5)dp->_bitstream[offset+1];
		dp->bitstream+=2;
		dp->nBitPos = 0;
	}

	int32D5 nVal = (int32D5)(((int64D5)dp->nBitBuff)>>(64-nBits));
	dp->nBitBuff <<= nBits;
	dp->nBitPos += nBits;

	if(dp->nBitPos>=32)
	{
		dp->nReadTotal++;
		//dp->nBitBuff += (((uint64D5)*dp->bitstream++)<<(dp->nBitPos-32));
		dp->nBitBuff += (((uint64D5)dp->_bitstream[offset])<<(dp->nBitPos-32));
		dp->nBitPos -= 32;
		dp->bitstream++;
	}

	return nVal;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Block Header Parser
///////////////////////////////////////////////////////////////////////////////////////////////////

int32D5 get_motion(int32D5 *mv, DecoderParam *dp, int32D5 x, int32D5 y, int32D5 pmv, char bBoundary)
{
	if(dp->nOptCode & OPTIMIZE_MV)
	{
		if(getbits(dp, 1))
			*mv = 3;
		else
			*mv = 7;
	}
	else 
	{
		// motion vector  
		if(getbits(dp, 1))
			*mv = pmv;
		else
			*mv = getbits(dp, 3);
	}

	if((y==0 || y==1) && (*mv<7))
	{
		printf("Syntax Error : Motion Vector is '%d' at (x %d, y %d). It should be 7 at top-line.\n", *mv, x, y);
		return ERROR_SYNTAX_MV;
	}

#ifdef DUMP_PARSE_LOG
	fprintf(fp_parse, "MV: %d\n", *mv);
#endif

	return SUCCESS;
}

void get_predictor(DecoderParam *dp, int32D5 x, int32D5 y, int32D5 mv, char bBoundary)
{
	pixel *pPred[3];
	GetOffset(dp->pRec, pPred, dp->nWidth, x, y);

	int32D5 disp[7] = {-4, -2, -1, 0, 1, 2, 4};
	int32D5 search [7][2] = {{-2, 0}, {-1, 0}, {-1, 1}, {0, 0}, {0, 1}, {1, 0}, {2, 0}};

	if(mv==7)
	{
		int32D5 i, j;

		for(j=0; j<3; j++)
		{
			for(i=0; i<dp->nBlockSize[j]; i++)
				pPred[j][i] = (bBoundary) ? dp->nBnd_init_val : pPred[j][-1];
		}
	}
	else
	{		
		GetReconPixel(dp, pPred, x, y, disp[mv]);
	}
}

int32D5 get_skip_flag(DecoderParam *dp)
{
	if(dp->nOptCode & OPTIMIZE_SKIP) 
		return 0;

	return getbits(dp, 1);
}

void get_errorbound(DecoderParam *dp, int32D5 *nErrorBound)
{
	int32D5 qp_val = getbits(dp, 2);

	switch(qp_val)
	{
	case 0:
		break;
	case 1:
		*nErrorBound -= dp->nInc;	// 2;
		break;
	case 2:
		*nErrorBound += dp->nInc;	// 2;
		break;
	case 3:
		*nErrorBound = getbits(dp, 12);
		break;
	}
}


int32D5 get_dynamic_range(DecoderParam *dp, int y, int32D5 nDynamicRange[][SUB_BLOCK_SIZE], int32D5 nPrevMode[3][2], int32D5 bSkip)
{
	//  line ۵ɶ ʱȭ 
	int i, j, flag[3][SUB_BLOCK_NUM];

	for(j=0; j<3; j++)
	{
		nDynamicRange[j][SUB_BLOCK_0] = 0;
		nDynamicRange[j][SUB_BLOCK_1] = 0;
	}

	if(bSkip)
		return SUCCESS;

	// 2Ʈ б
	for(j=0; j<3; j++)
	{
		if(dp->nBlockSize[j]>0) {
			flag[j][SUB_BLOCK_0] = getbits(dp, 2);
			flag[j][SUB_BLOCK_1] = getbits(dp, 2);
		}
	}

	// header 
	for(j=0; j<3; j++)
	{
		for(i=0; i<dp->nBlockSize[j]/SUB_BLOCK_SIZE; i++)
		{
			switch(flag[j][i])
			{
			case 0:
				nDynamicRange[j][i] = nPrevMode[j][DYNAMIC_DELAY];
				break;
			case 1:
				nDynamicRange[j][i] = nPrevMode[j][DYNAMIC_DELAY] + 1;
				break;
			case 2:
				nDynamicRange[j][i] = nPrevMode[j][DYNAMIC_DELAY] - 1;
				break;
			case 3:
				nDynamicRange[j][i] = (dp->nBitDepth<16) ? getbits(dp, 4) : getbits(dp, 5);
				break;
			}

			nPrevMode[j][0] = nPrevMode[j][1];
			nPrevMode[j][1] = nDynamicRange[j][i];

			if(nDynamicRange[j][i]>dp->nBitDepth+1)		// nBitDepth + SignBit
			{
				printf("Syntax Error : Dynamic Range is %d at (y %d) and over than threshold. \n", nDynamicRange[j][i], y);
				return ERROR_SYNTAX_DR;
			}

#ifdef DUMP_PARSE_LOG
			fprintf(fp_parse, "DR: %d\n", nDynamicRange[j][i]);
#endif
		}
	}

	return SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Processing Top Control
///////////////////////////////////////////////////////////////////////////////////////////////////

// Function: Decode 1 line
// return value : zero is SUCCESS, non-zero is ERROR

int32D5 CurrLineProcess_NL(DecoderParam *dp, int32D5 y)
{
	int32D5 nResult;
	int32D5 x, tx, c;
	int32D5 bBlockDynamicRangeSkip;
	int32D5 nPrevMV[2], nCurrMV;
	int32D5 nDynamicRange[3][SUB_BLOCK_SIZE];
	int32D5 nErrorBound = 0;						// QP (Quantization Parameter)
	int32D5 nPixelMax = (1<<dp->nBitDepth)-1;
	pixel *recon[3];
	static int32D5 nPrevMode[3][2] = {0};
	
	GetBlockSize(dp, y);

	// x : frame ü column ǥ
	// tx : sub tile  column ǥ

	for(x=0, tx=0; x<dp->nWidth[0]; x+=BLOCK_SIZE)
	{
		char bTileStart = (tx==0) ? 1:0;

#ifdef DUMP_PARSE_LOG
		fprintf(fp_parse, "BLOCK X: %d\n", x);
#endif

		// Ÿ    , MV ʱȭ
		if(bTileStart)
		{
			nPrevMV[0] = nPrevMV[1] = 7;

			for(c=0; c<3; c++) 
			{
				nPrevMode[c][0 /*delay*/] = (y==0 || y==1) ? 7 : 4;
				nPrevMode[c][1 /*delay*/] = (y==0 || y==1) ? 7 : 4;
			}
		}

		tx = ((tx+BLOCK_SIZE)>=dp->nTileWidth) ? 0 : tx + BLOCK_SIZE;


		// parse block header

#ifdef LRC_LOSSY_SUPPORT
			// rate control
		if( (x/BLOCK_SIZE)%dp->nNumBlkInRCUnit==0 && (dp->bQPIncluded) )
			get_errorbound(dp, &nErrorBound);
#endif

		if((nResult = get_motion(&nCurrMV, dp, x, y, nPrevMV[MV_DELAY], bTileStart))!=SUCCESS) 
			goto SYNTAX_ERROR;

		nPrevMV[0] = nPrevMV[1]; nPrevMV[1] = nCurrMV;
				
			// skip flag
		bBlockDynamicRangeSkip = get_skip_flag(dp);

			// dynamic range
		if((nResult = get_dynamic_range(dp, y, nDynamicRange, nPrevMode, bBlockDynamicRangeSkip))!=SUCCESS) 
			goto SYNTAX_ERROR;


		// parse block residue

		get_predictor(dp, x, y, nCurrMV, bTileStart);
		GetOffset(dp->pRec, recon, dp->nWidth, x, y);

		for(c=0; c<3; c++)
		{
			int i;

			for(i=0; i<dp->nBlockSize[c]/SUB_BLOCK_SIZE; i++)
			{
				residue px[SUB_BLOCK_SIZE];
				residue resi[SUB_BLOCK_SIZE] = {0,};

				// get residue

				if(nDynamicRange[c][i]>0)
				{
					resi[0] = getbits_sign(dp, nDynamicRange[c][i]);
					resi[1] = getbits_sign(dp, nDynamicRange[c][i]);
					resi[2] = getbits_sign(dp, nDynamicRange[c][i]);
					resi[3] = getbits_sign(dp, nDynamicRange[c][i]);

#ifdef LRC_LOSSY_SUPPORT
				// de-quantization
					resi[0] = resi[0] * (2*nErrorBound+1) + nErrorBound;
					resi[1] = resi[1] * (2*nErrorBound+1) + nErrorBound;
					resi[2] = resi[2] * (2*nErrorBound+1) + nErrorBound;
					resi[3] = resi[3] * (2*nErrorBound+1) + nErrorBound;
#endif
				}
#ifdef LRC_LOSSY_SUPPORT
				else
				{
				// de-quantization
					resi[0] = nErrorBound;
					resi[1] = nErrorBound;
					resi[2] = nErrorBound;
					resi[3] = nErrorBound;
				}
#endif

#ifdef DUMP_PARSE_LOG
				fprintf(fp_parse, "RES: %04x %04x %04x %04x\n", resi[0], resi[1], resi[2], resi[3]);
#endif

				// predictor + residue

				px[0] = recon[c][SUB_BLOCK_SIZE*i+0] + resi[0];
				px[1] = recon[c][SUB_BLOCK_SIZE*i+1] + resi[1];
				px[2] = recon[c][SUB_BLOCK_SIZE*i+2] + resi[2];
				px[3] = recon[c][SUB_BLOCK_SIZE*i+3] + resi[3];

				// clipping: 0 <= range <= nPixelMax

				recon[c][SUB_BLOCK_SIZE*i+0] = (pixel)CLIP(px[0], 0, nPixelMax);
				recon[c][SUB_BLOCK_SIZE*i+1] = (pixel)CLIP(px[1], 0, nPixelMax);
				recon[c][SUB_BLOCK_SIZE*i+2] = (pixel)CLIP(px[2], 0, nPixelMax);
				recon[c][SUB_BLOCK_SIZE*i+3] = (pixel)CLIP(px[3], 0, nPixelMax);
			}
		}
	}

	return SUCCESS;

SYNTAX_ERROR:

	return nResult;
}


// Function: Decode 1 frame
// return value : zero is SUCCESS, non-zero is ERROR

int LRC_Decode(DecoderParam *dp, unsigned char *pRecon, int *nWidth, int *nHeight) {

	int nResult;

	// Ķ  ޸ ʱȭ
	//DecoderParam dp; //static 
	//IntializeDecoderParam(&dp);
	//dp._bitstream = (uint32D5 *)pStream;
	//bitstream_init(&dp);

	// Image Header м
	if (!ReadImageHeader(dp))
		return ERROR_STREAM;		

	// Ķ   Memory ʱȭ
	InitDecoderSeq(dp);
	
	// ùο ƹ reference pixel Ƿ rec  0 ʱȭ س 
	int c;
	for(c=0; c<3; c++)	memset(dp->pRec[c], 0, sizeof(pixel)*dp->nWidth[c]);

#ifdef DUMP_PARSE_LOG
	fp_parse = fopen("parse.log", "w");
#endif

	// Vertical loop
	int h;
	for(h=0; h<dp->nHeight[0]; h++)
	{
		SetBitstreamPos(dp, h);			//   ù° bit  ̵

#ifdef DUMP_PARSE_LOG
		fprintf(fp_parse, "LINE: %d\n", h);
		if(h>100) break;
#endif

		nResult = CurrLineProcess_NL(dp, h);		// Line by Line Decoding

		if(nResult != SUCCESS) break;
	}

	//ReconFrameDump(&dp);
	ReconFormatConvert((pixel*)pRecon, dp);

	*nWidth = dp->nOriWidth;
	*nHeight = dp->nOriHeight;

#ifdef DUMP_PARSE_LOG
	fclose(fp_parse);
#endif

	// close memory
	CloseDecoderSeq(dp);

	return nResult;
}

//File IO
void FileIOInit(DecoderParam *dp, const char *InputAddr)
{
	int FileSize=0;
	
	// Memory allocation for file names
	// filename_srw : file name for input SRW file.
	// filename_rec : file name for output raw stream. 
	dp->ip.filename_srw = (char *)malloc(sizeof(char) * 256);
	if(dp->ip.filename_srw == NULL)
	{
		printf("dp->ip.filename_srw Memory allocation error...\n");
		exit(-1) ;
	}

	// copy the input file address to filenam_srw variable.	
	strcpy(dp->ip.filename_srw, InputAddr);

	// input SRW file open
	if((dp->ip.fp_srw = fopen(dp->ip.filename_srw, "rb"))==NULL)		
	{
		printf("%s file open error\n", dp->ip.filename_srw);
		exit(-1);		
	}	

	fseek(dp->ip.fp_srw, 0, SEEK_END);		
	FileSize = ftell(dp->ip.fp_srw);
	fseek(dp->ip.fp_srw, 0, SEEK_SET);

	if(!dp->srw_stream)
		dp->srw_stream	= (uint8D5 *)malloc(sizeof(uint8D5) * FileSize);	
	fread(dp->srw_stream, FileSize, 1, dp->ip.fp_srw);
}

void setMP(MetaDataParam *mp)
{
/*
PriRaw	SubfileType	off	00FE	M	N
PriRaw	ImageWidth	off	0100	M	N
PriRaw	ImageHeight	off	0101	M	N
PriRaw	BitsPerSample	off	0102	M	N
PriRaw	RawDataType	off	0103	M	N
PriRaw	StripOffsets	off	0111	M	N
PriRaw	SamplesPerPixel	off	0115	M	N
PriRaw	RowsPerStrip	off	0116	M	N
PriRaw	StripByteCounts	off	0117	M	N
PriRaw	CFARepeatPatternDim	off	828D	M	N
PriRaw	CFAPattern2	off	828E	M	N
PriRaw	CompMapDataOffset	off	A010	O	N
PriRaw	CompMapDataSize	off	A011	O	N
PriRaw	RawDataEndian	off	A101	O	N
PriRaw	RawDataPaddingInfo	off	A102	O	N
*/
/*
ZerothIFD	SubIFDOffset	off	014A	M	M
ZerothIFD	NextIFDOffset	off	0000	M	M
*/

	//Zeroth IFD
	strcpy(mp->reference[0].IFD_Name, "ZerothIFD"); strcpy(mp->reference[0].Tag_Name, "SubIFDOffset"); strcpy(mp->reference[0].Check, "off"); strcpy(mp->reference[0].Tag_ID, "014A"); strcpy(mp->reference[0].default_SRW, "M"); strcpy(mp->reference[0].default_JPG, "N");
	strcpy(mp->reference[1].IFD_Name, "ZerothIFD"); strcpy(mp->reference[1].Tag_Name, "NextIFDOffset"); strcpy(mp->reference[1].Check, "off"); strcpy(mp->reference[1].Tag_ID, "0000"); strcpy(mp->reference[1].default_SRW, "M"); strcpy(mp->reference[1].default_JPG, "N");

	//Primary Raw IFD
	strcpy(mp->reference[11].IFD_Name, "PriRaw"); strcpy(mp->reference[11].Tag_Name, "SubfileType"); strcpy(mp->reference[11].Check, "off"); strcpy(mp->reference[11].Tag_ID, "00FE"); strcpy(mp->reference[11].default_SRW, "M"); strcpy(mp->reference[11].default_JPG, "N");
	strcpy(mp->reference[12].IFD_Name, "PriRaw"); strcpy(mp->reference[12].Tag_Name, "ImageWidth"); strcpy(mp->reference[12].Check, "off"); strcpy(mp->reference[12].Tag_ID, "0100"); strcpy(mp->reference[12].default_SRW, "M"); strcpy(mp->reference[12].default_JPG, "N");
	strcpy(mp->reference[13].IFD_Name, "PriRaw"); strcpy(mp->reference[13].Tag_Name, "ImageHeight"); strcpy(mp->reference[13].Check, "off"); strcpy(mp->reference[13].Tag_ID, "0101"); strcpy(mp->reference[13].default_SRW, "M"); strcpy(mp->reference[13].default_JPG, "N");
	strcpy(mp->reference[14].IFD_Name, "PriRaw"); strcpy(mp->reference[14].Tag_Name, "BitsPerSample"); strcpy(mp->reference[14].Check, "off"); strcpy(mp->reference[14].Tag_ID, "0102"); strcpy(mp->reference[14].default_SRW, "M"); strcpy(mp->reference[14].default_JPG, "N");
	strcpy(mp->reference[15].IFD_Name, "PriRaw"); strcpy(mp->reference[15].Tag_Name, "RawDataType"); strcpy(mp->reference[15].Check, "off"); strcpy(mp->reference[15].Tag_ID, "0103"); strcpy(mp->reference[15].default_SRW, "M"); strcpy(mp->reference[15].default_JPG, "N");
	strcpy(mp->reference[16].IFD_Name, "PriRaw"); strcpy(mp->reference[16].Tag_Name, "StripOffsets"); strcpy(mp->reference[16].Check, "off"); strcpy(mp->reference[16].Tag_ID, "0111"); strcpy(mp->reference[16].default_SRW, "M"); strcpy(mp->reference[16].default_JPG, "N");
	strcpy(mp->reference[17].IFD_Name, "PriRaw"); strcpy(mp->reference[17].Tag_Name, "SamplesPerPixel"); strcpy(mp->reference[17].Check, "off"); strcpy(mp->reference[17].Tag_ID, "0115"); strcpy(mp->reference[17].default_SRW, "M"); strcpy(mp->reference[17].default_JPG, "N");
	strcpy(mp->reference[18].IFD_Name, "PriRaw"); strcpy(mp->reference[18].Tag_Name, "RowsPerStrip"); strcpy(mp->reference[18].Check, "off"); strcpy(mp->reference[18].Tag_ID, "0116"); strcpy(mp->reference[18].default_SRW, "M"); strcpy(mp->reference[18].default_JPG, "N");
	strcpy(mp->reference[19].IFD_Name, "PriRaw"); strcpy(mp->reference[19].Tag_Name, "StripByteCounts"); strcpy(mp->reference[19].Check, "off"); strcpy(mp->reference[19].Tag_ID, "0117"); strcpy(mp->reference[19].default_SRW, "M"); strcpy(mp->reference[19].default_JPG, "N");	
}